# GDB pretty printers for STLport.
#
# Copyright (C) 2008, 2009, 2010 Free Software Foundation, Inc.
# Copyright (C) 2010 Joachim Reichel
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# pylint: disable=C0103,C0111,R0201,R0903


import gdb
import re


# Set the STLport version which is needed for a few features.
#
# - for std::list:
#   STLport older than 5.0?
# - for std::deque, std::stack, and std::queue on 64bit systems:
#   STLport older than 5.2?
stlport_version = 5.2

# Indicates whether std::vector is printed with indices.
print_vector_with_indices = False


def lookup_stlport_type (typename):
    "Look up a type in the public STLport namespace."

    try: # STLport 5.0 and later, _STL_DEBUG defined
        return gdb.lookup_type ( 'stlpd_std::' + typename)
    except RuntimeError:
        try: # STLport 5.0 and later, _STL_DEBUG undefined
            return gdb.lookup_type ('stlp_std::' + typename)
        except RuntimeError: # STLport 4.6
            return gdb.lookup_type ('_STL::' + typename)


def lookup_stlport_priv_type (typename):
    "Look up a type in the private STLport namespace."

    try: # STLport 5.2, _STL_DEBUG defined
        return gdb.lookup_type ( 'stlpd_std::priv::' + typename)
    except RuntimeError:
        try: # STLport 5.2, _STL_DEBUG undefined
            return gdb.lookup_type ('stlp_std::priv::' + typename)
        except RuntimeError:
            try: # STLport 5.1
                return gdb.lookup_type ('stlp_priv::' + typename)
            except RuntimeError:
                try: # STLport 5.0, _STL_DEBUG defined
                    return gdb.lookup_type ('stlpd_std::' + typename)
                except RuntimeError:
                    try: # STLport 5.0, _STL_DEBUG undefined
                        return gdb.lookup_type ('stlp_std::' + typename)
                    except RuntimeError:
                        # STLport 4.6
                        return gdb.lookup_type ('_STL::' + typename)


def get_non_debug_impl (value, member = None):
    "Return the non-debug implementation of value or value[member]."
    if member:
        value = value[member]
    try:
        return value['_M_non_dbg_impl']
    except RuntimeError:
        return value


class RbtreeIterator:

    def __init__ (self, rbtree):
        tree = get_non_debug_impl (rbtree , '_M_t')
        self.size = tree['_M_node_count']
        self.node = tree['_M_header']['_M_data']['_M_left']
        self.count = 0

    def __iter__ (self):
        return self

    def __len__ (self):
        return int (self.size)

    def next (self):
        if self.count == self.size:
            raise StopIteration
        result = self.node
        self.count += 1
        if self.count < self.size:
            node = self.node
            # Is there a right child?
            if node.dereference()['_M_right']:
                # Walk down to left-most child in right subtree.
                node = node.dereference()['_M_right']
                while node.dereference()['_M_left']:
                    node = node.dereference()['_M_left']
            else:
                # Walk up to first parent reached via left subtree.
                parent = node.dereference()['_M_parent']
                while node == parent.dereference()['_M_right']:
                    node = parent
                    parent = parent.dereference()['_M_parent']
                node = parent
            self.node = node
        return result


class STLportBitsetPrinter:
    "Pretty printer for STLport std::bitset."

    def __init__(self, typename, val):
        self.typename = typename
        self.val      = val

    def to_string (self):
        # If template_argument handled values, we could print the
        # size.  Or we could use a regexp on the type.
        return '%s' % (self.typename)

    def children (self):
        words = self.val['_M_w']

        # The _M_w member can be either an unsigned long, or an
        # array.  This depends on the template specialization used.
        # If it is a single long, convert to a single element list.
        if words.type.code == gdb.TYPE_CODE_ARRAY:
            word_size = words.type.target ().sizeof
            n_words   = words.type.sizeof / word_size
        else:
            word_size = words.type.sizeof 
            n_words   = 1
            words     = [words]

        result = []
        word = 0
        while word < n_words:
            w = words[word]
            bit = 0
            while w != 0:
                if w & 1:
                    result.append (('[%d]' % (word * word_size * 8 + bit), 1))
                bit += 1
                w = w >> 1
            word += 1
        return result


class STLportDequePrinter:
    "Pretty printer for STLport std::deque."

    class Iterator:
        def __init__ (self, start_node, start_cur, start_last,
                      finish_cur, buffer_size):
            self.node        = start_node
            self.item        = start_cur
            self.node_last   = start_last
            self.last        = finish_cur
            self.buffer_size = buffer_size
            self.count       = 0

        def __iter__ (self):
            return self

        def next (self):
            if self.item == self.last:
                raise StopIteration
            result = ('[%d]' % self.count, self.item.dereference())
            self.count += 1
            self.item  += 1
            if self.item == self.node_last:
                self.node += 1
                self.item = self.node[0]
                self.node_last = self.item + self.buffer_size
            return result

    def __init__ (self, typename, val):
        self.typename = typename
        self.val = get_non_debug_impl (val)
        size = val.type.template_argument(0).sizeof
        # see MAX_BYTES in stlport/stl/_alloc.h
        if stlport_version < 5.2:
            blocksize = 128
        else:
            blocksize = 32 * gdb.lookup_type ("void").pointer().sizeof
        if size < blocksize:
            self.buffer_size = int (blocksize / size)
        else:
            self.buffer_size = 1

    def to_string (self):
        start   = self.val['_M_start']
        finish  = self.val['_M_finish']
        delta_n = finish['_M_node'] - start['_M_node'] - 1
        delta_s = start['_M_last'] - start['_M_cur']
        delta_f = finish['_M_cur'] - finish['_M_first']
        if delta_n == -1:
            size = delta_f
        else:
            size = self.buffer_size * delta_n + delta_s + delta_f
        ta0 = self.val.type.template_argument (0)
        return '%s<%s> with %d elements' % (self.typename, ta0, int (size))

    def children (self):
        start  = self.val['_M_start']
        finish = self.val['_M_finish']
        return self.Iterator (start['_M_node'], start['_M_cur'],
            start['_M_last'], finish['_M_cur'], self.buffer_size)

    def display_hint (self):
        return 'array'


class STLportListPrinter:
    "Pretty printer for STLport std::list."

    class Iterator:
        def __init__ (self, node_type, head):
            self.node_type = node_type
            # see list() in stlport/stl/_list.h
            if stlport_version < 5.0:
                self.sentinel = head
            else:
                self.sentinel = head.address
            self.item  = head['_M_next']
            self.count = 0

        def __iter__ (self):
            return self

        def next (self):
            if self.item == self.sentinel:
                raise StopIteration
            node = self.item.cast (self.node_type).dereference()
            self.item = node['_M_next']
            count = self.count
            self.count += 1
            return ('[%d]' % count, node['_M_data'])

    def __init__(self, typename, val):
        self.typename = typename
        self.val = get_non_debug_impl (val)

    def children (self):
        ta0       = self.val.type.template_argument(0)
        node_type = lookup_stlport_priv_type ('_List_node<%s>' % ta0).pointer()
        return self.Iterator (node_type, self.val['_M_node']['_M_data'])

    def to_string (self):
        ta0 = self.val.type.template_argument (0)
        if stlport_version < 5.0:
            sentinel = self.val['_M_node']['_M_data']
        else:
            sentinel = self.val['_M_node']['_M_data'].address
        if self.val['_M_node']['_M_data']['_M_next'] == sentinel:
            return 'empty %s<%s>' % (self.typename, ta0)
        return '%s<%s>' % (self.typename, ta0)

    def display_hint (self):
        return 'array'


class STLportMapPrinter:
    "Pretty printer for STLport std::map and std::multimap."

    class Iterator:

        def __init__ (self, rbiter, node_type):
            self.rbiter    = rbiter
            self.node_type = node_type
            self.count     = 0

        def __iter__ (self):
            return self

        def next (self):
            if self.count % 2 == 0:
                next_item = self.rbiter.next().dereference()
                self.pair = (next_item.cast (self.node_type))['_M_value_field']
                element = self.pair['first']
            else:
                element = self.pair['second']
            count = self.count
            self.count += 1
            return ('[%d]' % count, element)

    def __init__ (self, typename, val):
        self.typename = typename
        self.val = val

    def children (self):
        key_type   = self.val.type.template_argument (0)
        value_type = self.val.type.template_argument (1)
        pair_type  \
            = lookup_stlport_type ('pair<%s const,%s>' % (key_type,value_type))
        node_type  \
            = lookup_stlport_priv_type ('_Rb_tree_node<%s >' % str (pair_type))
        return self.Iterator (RbtreeIterator (self.val), node_type)

    def to_string (self):
        ta0 = self.val.type.template_argument (0)
        count = get_non_debug_impl (self.val, '_M_t')['_M_node_count']
        return ('%s<%s> with %d elements' % (self.typename, ta0, count))

    def display_hint (self):
        return 'map'


class STLportSetPrinter:
    "Pretty printer for STLport std::set and std::multiset."

    class Iterator:
        def __init__ (self, rbiter, node_type):
            self.rbiter    = rbiter
            self.node_type = node_type
            self.count     = 0

        def __iter__ (self):
            return self

        def next (self):
            next_item = self.rbiter.next().dereference()
            element = (next_item.cast (self.node_type))['_M_value_field']
            count = self.count
            self.count += 1
            return ('[%d]' % count, element)

    def __init__ (self, typename, val):
        self.typename = typename
        self.val = val

    def children (self):
        value_type = self.val.type.template_argument (0)
        node_type  \
            = lookup_stlport_priv_type ('_Rb_tree_node<%s>' % (value_type))
        return self.Iterator (RbtreeIterator (self.val), node_type)

    def to_string (self):
        ta0 = self.val.type.template_argument (0)
        count = get_non_debug_impl (self.val, '_M_t')['_M_node_count']
        return ('%s<%s> with %d elements' % (self.typename, ta0, count))

    def display_hint (self):
        return 'array'


class STLportSlistPrinter:
    "Pretty printer for STLport std::slist."

    class Iterator:
        def __init__ (self, node_type, head):
            self.node_type = node_type
            self.item  = head['_M_next']
            self.count = 0

        def __iter__ (self):
            return self

        def next (self):
            if self.item == 0:
                raise StopIteration
            node = self.item.cast (self.node_type).dereference()
            self.item = node['_M_next']
            count = self.count
            self.count += 1
            return ('[%d]' % count, node['_M_data'])

    def __init__(self, typename, val):
        self.typename = typename
        self.val = get_non_debug_impl (val)

    def children (self):
        ta0       = self.val.type.template_argument(0)
        node_type = lookup_stlport_priv_type ('_Slist_node<%s>' % ta0).pointer()
        return self.Iterator (node_type, self.val['_M_head']['_M_data'])

    def to_string (self):
        ta0 = self.val.type.template_argument (0)
        if self.val['_M_head']['_M_data']['_M_next'] == 0:
            return 'empty %s<%s>' % (self.typename, ta0)
        return '%s<%s>' % (self.typename, ta0)

    def display_hint (self):
        return 'array'


class STLportStringPrinter:
    "Pretty printer for STLport std::string or std::wstring."

    def __init__ (self, _typename, val):
        self.val = get_non_debug_impl (val)

    def to_string (self):
        try: # STLport 5.2 and later
            return self.val['_M_start_of_storage']['_M_data']
        except RuntimeError:
            try: # STLport 5.0 and 5.1
                static_buf = self.val['_M_buffers']['_M_static_buf']
                data       = self.val['_M_end_of_storage']['_M_data']
                if static_buf.address + 1 == data:
                    ta0 = self.val.type.template_argument (0)
                    return static_buf.cast (ta0.pointer())
                return self.val['_M_buffers']['_M_dynamic_buf']
            except RuntimeError: # STLport 4.6
                return self.val['_M_start']

    def display_hint (self):
        return 'string'


class STLportVectorPrinter:
    "Pretty printer for STLport std::vector."

    class Iterator:

        def __init__ (self, start, finish, bit_vector):
            self.bit_vector = bit_vector
            self.count      = 0
            if bit_vector:
                self.item   = start['_M_p']
                self.io     = start['_M_offset']
                self.finish = finish['_M_p']
                self.fo     = finish['_M_offset']
                self.isize  = 8 * self.item.dereference().type.sizeof
            else:
                self.item   = start
                self.finish = finish

        def __iter__ (self):
            return self

        def next (self):
            count = self.count
            self.count += 1
            if self.bit_vector:
                if self.item == self.finish and self.io == self.fo:
                    raise StopIteration
                element = self.item.dereference()
                value = 0
                if element & (1 << self.io):
                    value = 1
                self.io += 1
                if self.io >= self.isize:
                    self.item += 1
                    self.io   =  0
                return ('[%d]' % count, value)
            else:
                if self.item == self.finish:
                    raise StopIteration
                element = self.item.dereference()
                self.item += 1
                return ('[%d]' % count, element)

    def __init__ (self, typename, val):
        self.typename = typename
        self.val = get_non_debug_impl (val)
        self.bit_vector \
            = val.type.template_argument (0).code == gdb.TYPE_CODE_BOOL

    def children (self):
        start  = self.val['_M_start']
        finish = self.val['_M_finish']
        return self.Iterator (start, finish, self.bit_vector)

    def to_string (self):
        if self.bit_vector:
            start    = self.val['_M_start']['_M_p']
            so       = self.val['_M_start']['_M_offset']
            finish   = self.val['_M_finish']['_M_p']
            fo       = self.val['_M_finish']['_M_offset']
            end      = self.val['_M_end_of_storage']['_M_data']
            isize    = 8 * start.dereference().type.sizeof
            length   = (isize - so) + isize * (finish - start - 1) + fo
            capacity = isize * (end - start)
            return ('%s<bool> of length %d, capacity %d'
                % (self.typename, length, capacity))
        else:
            start    = self.val['_M_start']
            finish   = self.val['_M_finish']
            end      = self.val['_M_end_of_storage']['_M_data']
            length   = finish - start
            capacity = end - start
            ta0      = self.val.type.template_argument (0)
            return ('%s<%s> of length %d, capacity %d'
                % (self.typename, ta0, length, capacity))

    def display_hint (self):
        if print_vector_with_indices:
            return None
        else:
            return 'array'


class STLportWrapperPrinter:
    "Pretty printer for STLport std::stack, std::queue, and std::priority_queue."

    def __init__ (self, typename, val):
        self.typename = typename
        self.val = val
        self.visualizer = gdb.default_visualizer (val['c'])

    def children (self):
        return self.visualizer.children()

    def to_string (self):
        ta0 = self.val.type.template_argument (0)
        return ('%s<%s>, wrapping %s'
            % (self.typename, ta0, self.visualizer.to_string()))

    def display_hint (self):
        if hasattr (self.visualizer, 'display_hint'):
            return self.visualizer.display_hint()
        return None


def lookup_function (val):
    "Look-up and return a pretty-printer that can print val."

    type = val.type
    if type.code == gdb.TYPE_CODE_REF:
        type = type.target()
    type = type.unqualified().strip_typedefs()

    typename = type.tag
    if typename == None:
        return None

    for function in pretty_printers_dict:
        if function.search (typename):
            return pretty_printers_dict[function] (val)
    return None


def register_stlport_printers (obj):
    "Register STLport pretty-printers with object file obj."

    if obj == None:
        obj = gdb
    obj.pretty_printers.append (lookup_function)



pretty_printers_dict = {}

def add_entry (regex, printer, typename):
    prefix = "^(stlpd?_std|_STL)::"
    suffix = "<.*>$"
    pretty_printers_dict[re.compile (prefix+regex+suffix)] \
        = lambda val: printer (typename, val)

add_entry ("basic_string",   STLportStringPrinter,  None)
add_entry ("bitset",         STLportBitsetPrinter,  "std::bitset")
add_entry ("deque",          STLportDequePrinter,   "std::deque")
add_entry ("map",            STLportMapPrinter,     "std::map")
add_entry ("list",           STLportListPrinter,    "std::list")
add_entry ("multimap",       STLportMapPrinter,     "std::multimap")
add_entry ("multiset",       STLportSetPrinter,     "std::multiset")
add_entry ("queue",          STLportWrapperPrinter, "std::queue")
add_entry ("priority_queue", STLportWrapperPrinter, "std::priority_queue")
add_entry ("set",            STLportSetPrinter,     "std::set")
add_entry ("slist",          STLportSlistPrinter,   "std::slist")
add_entry ("stack",          STLportWrapperPrinter, "std::stack")
add_entry ("vector",         STLportVectorPrinter,  "std::vector")
